EDNA_TEST_DATA = """<?xml version="1.0" ?>
<XSDataInputMXCuBE>
  <dataCollectionId>
    <value>976533</value>
  </dataCollectionId>
  <diffractionPlan>
    <aimedIOverSigmaAtHighestResolution>
      <value>3.000000e+00</value>
    </aimedIOverSigmaAtHighestResolution>
    <complexity>
      <value>none</value>
    </complexity>
    <maxExposureTimePerDataCollection>
      <value>6.000000e+03</value>
    </maxExposureTimePerDataCollection>
  </diffractionPlan>
  <experimentalCondition>
    <beam>
      <flux>
        <value>2.710000e+12</value>
      </flux>
      <minExposureTimePerImage>
        <value>8.000000e-02</value>
      </minExposureTimePerImage>
      <size>
        <x>
          <value>5.000000e-02</value>
        </x>
        <y>
          <value>3.000000e-02</value>
        </y>
      </size>
      <transmission>
        <value>1.000000e+02</value>
      </transmission>
      <wavelength>
        <value>9.136656e-01</value>
      </wavelength>
    </beam>
    <goniostat>
      <maxOscillationSpeed>
        <value>3.600000e+02</value>
      </maxOscillationSpeed>
      <minOscillationWidth>
        <value>5.000000e-02</value>
      </minOscillationWidth>
    </goniostat>
  </experimentalCondition>
  <sample>
    <shape>
      <value>1.000000e+00</value>
    </shape>
    <size>
      <x>
        <value>1.000000e-01</value>
      </x>
      <y>
        <value>1.000000e-01</value>
      </y>
      <z>
        <value>1.000000e-01</value>
      </z>
    </size>
    <susceptibility>
      <value>1.000000e+00</value>
    </susceptibility>
  </sample>
  <dataSet>
    <imageFile>
      <path>
        <value>/data/id29/inhouse/DAS/RAW_DATA/ref-4ESR372D11_1_0001.cbf</value>
      </path>
    </imageFile>
    <imageFile>
      <path>
        <value>/data/id29/inhouse/DAS/RAW_DATA/ref-4ESR372D11_1_0002.cbf</value>
      </path>
    </imageFile>
    <imageFile>
      <path>
        <value>/data/id29/inhouse/DAS/RAW_DATA/ref-4ESR372D11_1_0003.cbf</value>
      </path>
    </imageFile>
    <imageFile>
      <path>
        <value>/data/id29/inhouse/DAS/RAW_DATA/ref-4ESR372D11_1_0004.cbf</value>
      </path>
    </imageFile>
  </dataSet>
</XSDataInputMXCuBE>
"""

EDNA_DEFAULT_INPUT = """<?xml version="1.0" ?>
<XSDataInputMXCuBE>
    <diffractionPlan>
        <aimedCompleteness>
            <value>9.900000e-01</value>
        </aimedCompleteness>
        <aimedIOverSigmaAtHighestResolution>
            <value>3.000000e+00</value>
        </aimedIOverSigmaAtHighestResolution>
        <aimedResolution>
            <value>1.000000e+00</value>
        </aimedResolution>
        <aimedMultiplicity>
            <value>4.000000e+00</value>
        </aimedMultiplicity>
        <complexity>
            <value>min</value>
        </complexity>
        <maxExposureTimePerDataCollection>
            <value>6.000000e+03</value>
        </maxExposureTimePerDataCollection>
    </diffractionPlan>
    <experimentalCondition>
        <beam>
            <minExposureTimePerImage>
                <value>1.000000e-01</value>
            </minExposureTimePerImage>
            <size>
                <x>
                    <value>1.000000e-01</value>
                </x>
                <y>
                    <value>0.70000e-01</value>
                </y>
            </size>
        </beam>
        <goniostat>
            <maxOscillationSpeed>
                <value>3.600000e+02</value>
            </maxOscillationSpeed>
            <minOscillationWidth>
                <value>2.000000e-01</value>
            </minOscillationWidth>
        </goniostat>
    </experimentalCondition>
    <sample>
        <size>
            <x>
                <value>1.000000e-01</value>
            </x>
            <y>
                <value>1.000000e-01</value>
            </y>
            <z>
                <value>1.000000e-01</value>
            </z>
        </size>
        <susceptibility>
            <value>1.000000e+00</value>
        </susceptibility>
    </sample>
</XSDataInputMXCuBE>
"""

EDNA_RESULT_DATA = """<?xml version="1.0" ?>
<XSDataResultMXCuBE>
    <characterisationResult>
        <dataCollection>
            <diffractionPlan>
                <aimedCompleteness>
                    <value>9.900000e-01</value>
                </aimedCompleteness>
                <aimedIOverSigmaAtHighestResolution>
                    <value>2.000000e+00</value>
                </aimedIOverSigmaAtHighestResolution>
                <aimedResolution>
                    <value>5.000000e-01</value>
                </aimedResolution>
                <anomalousData>
                    <value>false</value>
                </anomalousData>
                <complexity>
                    <value>none</value>
                </complexity>
                <detectorDistanceMax>
                    <value>1.300000e+03</value>
                </detectorDistanceMax>
                <detectorDistanceMin>
                    <value>1.350000e+02</value>
                </detectorDistanceMin>
                <forcedSpaceGroup>
                    <value></value>
                </forcedSpaceGroup>
                <maxExposureTimePerDataCollection>
                    <value>6.000000e+02</value>
                </maxExposureTimePerDataCollection>
            </diffractionPlan>
            <subWedge>
                <experimentalCondition>
                    <beam>
                        <exposureTime>
                            <value>4.000000e-02</value>
                        </exposureTime>
                        <flux>
                            <value>2.300000e+12</value>
                        </flux>
                        <minExposureTimePerImage>
                            <value>4.000000e-02</value>
                        </minExposureTimePerImage>
                        <size>
                            <x>
                                <value>5.000000e-02</value>
                            </x>
                            <y>
                                <value>5.000000e-02</value>
                            </y>
                        </size>
                        <transmission>
                            <value>1.000000e+02</value>
                        </transmission>
                        <wavelength>
                            <value>9.762460e-01</value>
                        </wavelength>
                    </beam>
                    <detector>
                        <beamPositionX>
                            <value>2.159546e+02</value>
                        </beamPositionX>
                        <beamPositionY>
                            <value>2.103680e+02</value>
                        </beamPositionY>
                        <distance>
                            <value>3.920700e+02</value>
                        </distance>
                        <name>
                            <value>PILATUS 6M</value>
                        </name>
                        <numberPixelX>
                            <value>2463</value>
                        </numberPixelX>
                        <numberPixelY>
                            <value>2527</value>
                        </numberPixelY>
                        <pixelSizeX>
                            <value>1.720000e-01</value>
                        </pixelSizeX>
                        <pixelSizeY>
                            <value>1.720000e-01</value>
                        </pixelSizeY>
                        <serialNumber>
                            <value>PILATUS 6M-F, S/N 60-0117-F</value>
                        </serialNumber>
                        <type>
                            <value>pilatus6m</value>
                        </type>
                    </detector>
                    <goniostat>
                        <maxOscillationSpeed>
                            <value>1.190000e+02</value>
                        </maxOscillationSpeed>
                        <minOscillationWidth>
                            <value>5.000000e-02</value>
                        </minOscillationWidth>
                        <oscillationWidth>
                            <value>1.000000e+00</value>
                        </oscillationWidth>
                        <rotationAxisEnd>
                            <value>9.100000e+01</value>
                        </rotationAxisEnd>
                        <rotationAxisStart>
                            <value>9.000000e+01</value>
                        </rotationAxisStart>
                    </goniostat>
                </experimentalCondition>
                <subWedgeNumber>
                    <value>1</value>
                </subWedgeNumber>
            </subWedge>
            <subWedge>
                <experimentalCondition>
                    <beam>
                        <exposureTime>
                            <value>4.000000e-02</value>
                        </exposureTime>
                        <flux>
                            <value>2.300000e+12</value>
                        </flux>
                        <minExposureTimePerImage>
                            <value>4.000000e-02</value>
                        </minExposureTimePerImage>
                        <size>
                            <x>
                                <value>5.000000e-02</value>
                            </x>
                            <y>
                                <value>5.000000e-02</value>
                            </y>
                        </size>
                        <transmission>
                            <value>1.000000e+02</value>
                        </transmission>
                        <wavelength>
                            <value>9.762460e-01</value>
                        </wavelength>
                    </beam>
                    <goniostat>
                        <maxOscillationSpeed>
                            <value>1.190000e+02</value>
                        </maxOscillationSpeed>
                        <minOscillationWidth>
                            <value>5.000000e-02</value>
                        </minOscillationWidth>
                        <oscillationWidth>
                            <value>1.000000e+00</value>
                        </oscillationWidth>
                        <rotationAxisEnd>
                            <value>1.810000e+02</value>
                        </rotationAxisEnd>
                        <rotationAxisStart>
                            <value>1.800000e+02</value>
                        </rotationAxisStart>
                    </goniostat>
                </experimentalCondition>
                <image>
                    <path>
                    </path>
                    <number>
                        <value>2</value>
                    </number>
                </image>
                <subWedgeNumber>
                    <value>2</value>
                </subWedgeNumber>
            </subWedge>
        </dataCollection>
        <strategyResult>
            <collectionPlan>
                <collectionPlanNumber>
                    <value>1</value>
                </collectionPlanNumber>
                <collectionStrategy>
                    <subWedge>
                        <experimentalCondition>
                            <beam>
                                <exposureTime>
                                    <value>4.000000e-02</value>
                                </exposureTime>
                                <flux>
                                    <value>2.300000e+12</value>
                                </flux>
                                <minExposureTimePerImage>
                                    <value>4.000000e-02</value>
                                </minExposureTimePerImage>
                                <size>
                                    <x>
                                        <value>5.000000e-02</value>
                                    </x>
                                    <y>
                                        <value>5.000000e-02</value>
                                    </y>
                                </size>
                                <transmission>
                                    <value>3.136190e+01</value>
                                </transmission>
                                <wavelength>
                                    <value>9.762460e-01</value>
                                </wavelength>
                            </beam>
                            <detector>
                                <beamPositionX>
                                    <value>2.156012e+02</value>
                                </beamPositionX>
                                <beamPositionY>
                                    <value>2.115154e+02</value>
                                </beamPositionY>
                                <distance>
                                    <value>3.593500e+02</value>
                                </distance>
                                <name>
                                    <value>PILATUS 6M</value>
                                </name>
                                <numberPixelX>
                                    <value>2463</value>
                                </numberPixelX>
                                <numberPixelY>
                                    <value>2527</value>
                                </numberPixelY>
                                <pixelSizeX>
                                    <value>1.720000e-01</value>
                                </pixelSizeX>
                                <pixelSizeY>
                                    <value>1.720000e-01</value>
                                </pixelSizeY>
                                <serialNumber>
                                    <value>PILATUS 6M-F</value>
                                </serialNumber>
                                <type>
                                    <value>pilatus6m</value>
                                </type>
                            </detector>
                            <goniostat>
                                <maxOscillationSpeed>
                                    <value>1.190000e+02</value>
                                </maxOscillationSpeed>
                                <minOscillationWidth>
                                    <value>5.000000e-02</value>
                                </minOscillationWidth>
                                <oscillationWidth>
                                    <value>1.000000e-01</value>
                                </oscillationWidth>
                                <rotationAxisEnd>
                                    <value>2.420000e+02</value>
                                </rotationAxisEnd>
                                <rotationAxisStart>
                                    <value>1.380000e+02</value>
                                </rotationAxisStart>
                            </goniostat>
                        </experimentalCondition>
                        <subWedgeNumber>
                            <value>1</value>
                        </subWedgeNumber>
                    </subWedge>
                </collectionStrategy>
                <strategySummary>
                    <completeness>
                        <value>1.000000e+00</value>
                    </completeness>
                    <iSigma>
                        <value>2.000000e+00</value>
                    </iSigma>
                    <rankingResolution>
                        <value>1.810000e+00</value>
                    </rankingResolution>
                    <redundancy>
                        <value>4.440000e+00</value>
                    </redundancy>
                    <resolution>
                        <value>1.810000e+00</value>
                    </resolution>
                    <resolutionReasoning>
                        <value>Resolution limit is set by the radiation damage</value>
                    </resolutionReasoning>
                    <totalDataCollectionTime>
                        <value>4.409600e+01</value>
                    </totalDataCollectionTime>
                    <totalExposureTime>
                        <value>72</value>
                    </totalExposureTime>
                </strategySummary>
            </collectionPlan>
        </strategyResult>
    </characterisationResult>
    <htmlPage>
        <path>
        </path>
    </htmlPage>
    <screeningId>
        <value>21514</value>
    </screeningId>
</XSDataResultMXCuBE>
"""
